import re
import os
import zfit
import pprint
import zutils.utils      as zut
import utils_noroot      as utnr
import hqm.model         as hqm_model
import matplotlib.pyplot as plt

from importlib.resources import files
from rkex_model          import model
from zutils.plot         import plot                     as zfp
from hqm.tools.scales    import scales                   as scl
from stats.average       import average                  as stav
from logzero             import logger                   as log
from builder             import builder                  as cb_builder
from misID_tools.zmodel  import misID_real_model_builder as msid

#---------------------------------------------------------------
class rk_model(model):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._d_val = {} 
        self._d_var = {} 

        self._nsig_ee    = None
        self._nsig_mm    = None
        self._read_yields= True
    #---------------------------------------------------------------
    @property
    def read_yields(self):
        return self._read_yields

    @read_yields.setter
    def read_yields(self, value):
        '''
        If true, will try to read yields from JSON files. This is needed to make toys
        not for actual fit to data
        '''
        self._read_yields= value
    #---------------------------------------------------------------
    def _preffix_to_info(self, preffix):
        regex = '(ee|mm)_(r1|r2p1|2017|2018|all)_(TOS|TIS)_.*'
        mtch  = re.match(regex, preffix)
        try:
            [chan, dset, trg] = mtch.groups()
        except:
            log.error(f'Cannot extract dataset and trigger from: {preffix}')
            raise

        trig = {'mm_TOS' : 'MTOS', 'mm_TIS' : 'MTIS', 'ee_TOS' : 'ETOS', 'ee_TIS' : 'GTIS'} [f'{chan}_{trg}']

        return dset, trig
    #---------------------------------------------------------------
    def _add_constraints(self, d_cns):
        d_val = { name : cns[0]    for name, cns in d_cns.items() }
        d_var = { name : cns[1]**2 for name, cns in d_cns.items() }

        self._d_val.update(d_val)
        self._d_var.update(d_var)
    #---------------------------------------------------------------
    def _fix_shape(self, pdf):
        s_par = pdf.get_params(is_yield=False)
        for par in s_par:
            par.floating=False

        return pdf
    #---------------------------------------------------------------
    def _get_combinatorial(self, preffix, fix_shape=False):
        dset, trig    = self._preffix_to_info(preffix) 
        obj           = cb_builder(dset=dset, trigger=trig, vers='v5', q2bin='high', const=False)
        obj.cache_path= './cb_buider.tar.gz'
        cmb, d_cns    = obj.get_pdf(obs=self._obs, preffix=f'cb_{preffix}', name='Combinatorial') 

        if fix_shape:
            cmb = self._fix_shape(cmb)
        else:
            self._add_constraints(d_cns)

        name  = r'$\mu\mu$' if preffix.startswith('mm_') else 'ee'
        nent  = self._get_entries(nsig=None, kind='cmb', trigger=trig, year=dset)
        ncb   = zfit.Parameter(f'ncb_{preffix}', nent, 0, 100000)
        cmb.set_yield(ncb)

        return cmb 
    #---------------------------------------------------------------
    def _get_rare_scale(self, year, trigger, kind):
        if   year == 'r1':
            obj_1        = scl(dset='2011', trig=trigger, kind=kind)
            val_1, err_1 = obj_1.get_scale()

            obj_2        = scl(dset='2012', trig=trigger, kind=kind)
            val_2, err_2 = obj_2.get_scale()
        elif year == 'r2p1':
            obj_1        = scl(dset='2015', trig=trigger, kind=kind)
            val_1, err_1 = obj_1.get_scale()

            obj_2        = scl(dset='2016', trig=trigger, kind=kind)
            val_2, err_2 = obj_2.get_scale()
        else:
            obj          = scl(dset=  year, trig=trigger, kind=kind)
            val, err     = obj.get_scale()

            return val, err

        val, err, _ = stav([val_1, val_2], [err_1, err_2])

        return val, err
    #---------------------------------------------------------------
    def _get_entries_from_json(self, kind=None, trig=None, year=None):
        if not self._read_yields:
            log.warning(f'Not reading yields from JSON for {kind}/{trig}/{year}, using zero events')
            return 0

        json_path = files('extractor_data').joinpath(f'sb_fits/{self._norm_vers}/{year}_{trig}.json')
        if not os.path.isfile(json_path):
            log.error(f'Cannot find: {json_path}')
            raise FileNotFoundError

        d_par = utnr.load_json(json_path)
        for key, [val, _] in d_par.items():
            flg_1 = key.startswith('ncb_') and kind == 'cmb'
            flg_2 = key.startswith('npr_') and kind == 'prc'
            if flg_1 or flg_2:
                return val

        log.error(f'Cannot find yield for {kind} among:')
        pprint.pprint(d_par)
        raise
    #---------------------------------------------------------------
    def _get_entries(self, nsig=None, kind=None, trigger=None, year=None):
        if   kind in ['prc', 'cmb']:
            nent  = self._get_entries_from_json(kind=kind, trig=trigger, year=year)

            log.info(f'Taking {kind} yield from JSON as: {nent:.0f}')
        elif kind in ['bpks', 'bdks', 'bsph']:
            scale = self._get_rare_scale(year, trigger, kind)
            scl   = scale[0]
            nent  = nsig * scl

            log.info(f'Taking {kind} yield as {nent:.0f}={scl:.3f} * {nsig:.0f}')
        else:
            log.error(f'Invalid kind: {kind}')
            raise

        return nent
    #---------------------------------------------------------------
    def _get_signal(self, preffix, nent=None):
        year, trig = self._preffix_to_info(preffix)
        log.warning(f'Using 2018 signal for {year}')
        year = '2018'
        sig, d_cns = hqm_model.get_signal_shape(preffix=preffix, year=year, trigger=trig)

        self._add_constraints(d_cns)

        nsg    = zfit.Parameter(f'nsg_{preffix}', nent, 0, 100000)
        esig   = sig.create_extended(nsg, name='Signal')

        if   preffix.startswith('ee_'):
            self._nsig_ee = nsg 
        elif preffix.startswith('mm_'):
            self._nsig_mm = nsg 
        else:
            log.error(f'Invalid preffix: {preffix}')
            raise ValueError

        return esig
    #---------------------------------------------------------------
    def _get_rare(self, preffix, nent=None, kind=None):
        year, trig = self._preffix_to_info(preffix)
        log.warning(f'Using 2018 rare {kind} background for {year}')
        year = '2018'

        if   kind == 'bpks':
            pdf = hqm_model.get_Bu2Ksee_shape(name=r'$B^+\to K^{*0}ee$', year=year, trigger=trig)
        elif kind == 'bdks':
            pdf = hqm_model.get_Bd2Ksee_shape(name=r'$B_0\to K^{*+}ee$', year=year, trigger=trig)
        elif kind == 'bsph':
            pdf = hqm_model.get_Bd2Ksee_shape(name=r'$B_s\to \phi  ee$', year=year, trigger=trig)
        else:
            log.error(f'Invalid kind: {kind}')
            raise ValueError

        nrare   = self._get_entries(nent, kind, trigger=trig, year=year)
        ns_o_nr = self._nsig_ee.value() / nrare 
        nbkg    = zfit.ComposedParameter(f'nr{kind}_{preffix}', lambda ns : ns / ns_o_nr, params=[self._nsig_ee])
        pdf.set_yield(nbkg)

        return pdf 
    #---------------------------------------------------------------
    def _get_prec(self, preffix):
        year, trig = self._preffix_to_info(preffix)
        nent= self._get_entries(nsig=None, kind='prc', trigger=trig, year=year)

        log.warning(f'Using 2018 PRec background for {year}')
        year = '2018'
        pdf = hqm_model.get_part_reco(name='$PRec + \psi(2S) K^+$', preffix=preffix, year=year, trigger=trig)
        npr = zfit.Parameter(f'npr_{preffix}', nent, 0, 100000)
        pdf.set_yield(npr)

        return pdf 
    #---------------------------------------------------------------
    def _get_msid(self, preffix):
        dset, trig    = self._preffix_to_info(preffix)
        bld= msid(name='Mis-ID', version=self._msid_vers, obs=self._obs, preffix=preffix)
        bld.load_model(dset, trig)
        pdf=bld.build_model() 

        return pdf
    #---------------------------------------------------------------
    def _add_combinatorial(self, l_pdf, preffix):
        _, trig   = self._preffix_to_info(preffix)

        if   trig == 'ETOS' and self._kind != 'cmb_ee:use_etos':
            return l_pdf

        cbkg = self._get_combinatorial(preffix, fix_shape= trig=='ETOS')
        l_pdf.insert(0, cbkg)

        return l_pdf
    #---------------------------------------------------------------
    def _get_pdf(self, preffix='', nent=None):
        preffix   = f'{preffix}_{self._preffix}'
        _, trig   = self._preffix_to_info(preffix)
        if trig == 'MTIS':
            return None

        esig      = self._get_signal(preffix, nent=nent)
        self._obs = esig.space

        if   preffix.startswith('ee_'):
            erbp  = self._get_rare(preffix, nent=nent, kind='bpks')
            erbd  = self._get_rare(preffix, nent=nent, kind='bdks')
            erbs  = self._get_rare(preffix, nent=nent, kind='bsph')
            eprc  = self._get_prec(preffix)
            emis  = self._get_msid(preffix)

            l_pdf = [erbp, erbd, erbs, eprc, emis, esig]
        elif preffix.startswith('mm_'):
            l_pdf = [esig]
        else:
            log.error(f'Invalid preffix: {preffix}')
            raise

        l_pdf = self._add_combinatorial(l_pdf, preffix)
        pdf   = l_pdf[0] if len(l_pdf) == 1 else zfit.pdf.SumPDF(l_pdf) 

        return pdf 
    #---------------------------------------------------------------
    def _plot_model(self, key, mod):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots/models'
        os.makedirs(plt_dir, exist_ok=True)

        obj= zfp(data=mod.create_sampler(fixed_params=False), model=mod)
        obj.plot(nbins=50, ext_text=key, stacked=True)

        log.info(f'Saving to: {plt_dir}/{key}.png')
        plt.savefig(f'{plt_dir}/{key}.png')
        plt.close('all')

        d_const = {}
        for name in self._d_val:
            val = self._d_val[name]
            var = self._d_var[name]
            d_const[name] = [val, var]

        log.info(f'Saving to: {plt_dir}/{key}.txt')
        zut.print_pdf(mod, d_const=d_const, txt_path=f'{plt_dir}/{key}.txt')
    #---------------------------------------------------------------
    def get_cons(self):
        '''
        Will return constraints on model parameters 

        Returns
        -----------
        d_val, d_var: Tuple of dictionaries pairing parameter name with value (mu of Gaussian) and 
        variance respectively.
        '''
        self._initialize()

        log.debug('-' * 20)
        log.debug(f'{"Name":<40}{"Value":<15}{"Variance":<15}')
        log.debug('-' * 20)
        for name in self._d_val:
            val = self._d_val[name]
            var = self._d_var[name]

            log.debug(f'{name:<40}{val:<15.3f}{var:<15.3f}')
        log.debug('-' * 20)

        return self._d_val, self._d_var
#---------------------------------------------------------------




import os
import re
import glob
import ROOT
import tqdm
import math
import numpy
import pprint
import pandas            as pnd
import utils_noroot      as utnr
import seaborn           as sns
import matplotlib.pyplot as plt
import read_selection    as rs

from importlib.resources import files
from scipy.interpolate   import griddata as spy_grid
from logzero             import logger   as log

#------------------------------------------------------------
class scale_maker:
    '''
    Class used to get and store efficiencies under different BDT working points
    '''
    #--------------------------
    def __init__(self, wp=None, step_size=None, dset=None, trig=None):
        '''
        Parameters
        --------------------------
        wp (dict): Stores working point ends for the scan, i.e. {'BDT_cmb' : (0.5, 1.0)} 
        step_size (float): Size of step to scan BDT score, e.g. 1e-2

        dset (str): Dataset, e.g. r1
        trig (str): Trigger, e.g. MTOS
        '''
        self._d_wp        = wp 
        self._step_size   = step_size
        self._dset        = dset
        self._trig        = trig

        self._xvar        = None
        self._yvar        = None
        self._cas_dir     = None
        self._out_dir     = None
        self._plt_dir     = None
        self._df_dat      = None
        self._size        = None
        self._version     = None
        self._proc        = 'sign'
        self._vers        = 'v10.21p2'
        self._jdir        = 'bdt_scales'

        self._initialized = False 
    #--------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._cas_dir = os.environ['CASDIR']
        self._check_wp()

        low_x, hig_x  = self._d_wp[self._xvar]
        low_y, hig_y  = self._d_wp[self._yvar]
        self._nstep_x = int((hig_x - low_x) / self._step_size)
        self._nstep_y = int((hig_y - low_y) / self._step_size)

        self._out_dir = files('extractor_data').joinpath(f'bdt_eff/{self._version}')
        os.makedirs(self._out_dir, exist_ok=True)

        self._plt_dir = f'{self._out_dir}/{self._dset}_{self._trig}' 
        os.makedirs(self._plt_dir, exist_ok=True)

        self._initialized = True
    #--------------------------
    def _check_wp(self):
        if len(self._d_wp) != 2:
            log.error(f'Not found only 2 variable in d_wp:')
            pprint.pprint(self._d_wp)
            raise

        self._xvar, self._yvar = self._d_wp.keys()
    #--------------------------
    def _get_data_paths(self):
        if   self._dset == 'r1':
            dat_dir_1 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2011_{self._trig}'
            dat_dir_2 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2012_{self._trig}'
            l_dat_dir = [dat_dir_1, dat_dir_2]
        elif self._dset == 'r2p1':
            dat_dir_1 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2015_{self._trig}'
            dat_dir_2 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2016_{self._trig}'
            l_dat_dir = [dat_dir_1, dat_dir_2]
        elif self._dset == 'all':
            dat_dir_1 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2011_{self._trig}'
            dat_dir_2 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2012_{self._trig}'
            dat_dir_3 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2015_{self._trig}'
            dat_dir_4 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2016_{self._trig}'
            dat_dir_5 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2017_{self._trig}'
            dat_dir_6 = f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/2018_{self._trig}'
            l_dat_dir = [dat_dir_1, dat_dir_2, dat_dir_3, dat_dir_4, dat_dir_5, dat_dir_6]
        else:
            l_dat_dir = [f'{self._cas_dir}/tools/apply_selection/{self._jdir}/{self._proc}/{self._vers}/{self._dset}_{self._trig}']

        l_l_path = [ glob.glob(f'{dat_dir}/*.root') for dat_dir in l_dat_dir ] 
        l_fpath  = [ fpath for l_path in l_l_path for fpath in l_path ]

        if len(l_fpath) == 0:
            log.error(f'No ROOT file found in: {l_dat_dir}')
            raise

        log.info(f'Picking up data from: {l_dat_dir}')

        return l_fpath
    #--------------------------
    def _get_data(self):
        l_fpath      = self._get_data_paths()
        rdf          = ROOT.RDataFrame(self._trig, l_fpath)
        rdf          = rdf.Define('year', 'int(yearLabbel)')

        l_var        = [self._xvar, self._yvar, 'year']
        d_arr        = { var : rdf.AsNumpy(l_var)[var].tolist() for var in l_var} 
        df           = pnd.DataFrame(d_arr) 
        df['weight'] = df.year.apply(self._weight_from_year)

        for var in df.columns:
            self._plot_var(var, df)

        log.info(f'Picking {self._size} entries with {l_var} vars from {len(l_fpath)} files from {self._trig}')


        self._size = df.weight.sum() 

        return df
    #--------------------------
    def _weight_from_year(self, year):
        d_lumi = {2011 : 1, 2012 : 2, 2015 : 0.3, 2016 : 1.6, 2017 : 1.7, 2018 : 2.1}

        if year not in d_lumi:
            log.error(f'Year {year} not valid')
            raise

        return d_lumi[year]
    #--------------------------
    def _get_efficiency(self, xval=None, yval=None):
        xvar, yvar = self._xvar, self._yvar

        df = self._df_dat
        df = df[df[xvar] > xval]
        df = df[df[yvar] > yval]

        npas = df.weight.sum() 
        eff  = npas / self._size

        if not (0 <= eff <= 1):
            log.error(f'Invalid efficiency: {eff:.3f} = {npas}/{self._size}')
            raise

        return eff
    #--------------------------
    def _get_efficiencies(self):
        low_x, hig_x = self._d_wp[self._xvar]
        low_y, hig_y = self._d_wp[self._yvar]

        l_xval = numpy.linspace(low_x, hig_x, num=int(self._nstep_x))
        l_yval = numpy.linspace(low_y, hig_y, num=int(self._nstep_y))

        l_xgrid= []
        l_ygrid= []
        l_zgrid= []
        for xval in tqdm.tqdm(l_xval, ascii=' -'):
            for yval in l_yval:
                zval = self._get_efficiency(xval=xval, yval=yval)
                l_xgrid.append(xval)
                l_ygrid.append(yval)
                l_zgrid.append(zval)

        df = pnd.DataFrame({self._xvar : l_xgrid, self._yvar : l_ygrid, 'Efficiency' : l_zgrid})

        return df 
    #--------------------------
    def _plot_var(self, varname, df):
        if self._out_dir is None:
            return

        plot_path = f'{self._plt_dir}/{varname}.png'

        if varname != 'weight':
            plt.hist(df[varname], weights=df.weight, bins=30)
        else:
            plt.hist(df[varname],                     bins=30)

        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------
    def _plot_eff(self, df, interpolate=None):
        if not self._out_dir:
            return

        fig = plt.figure(figsize=(15, 10))
        if interpolate:
            self._plot_interpolated(df)
        else:
            self._plot_true(df)

        plt.colorbar()
        plot_path = f'{self._plt_dir}/eff_int.png' if interpolate else f'{self._plt_dir}/eff_raw.png'
        log.info(f'Saving to: {plot_path}')
        plt.xlabel(self._xvar)
        plt.ylabel(self._yvar)
        plt.title(f'{self._dset}, {self._trig}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------
    def _plot_interpolated(self, df):
        low_x, hig_x   = self._d_wp[self._xvar]
        low_y, hig_y   = self._d_wp[self._yvar]
        grid_x, grid_y = numpy.mgrid[low_x:hig_x:1000j, low_y:hig_y:1000j]
        
        x_arr, y_arr   = df[self._xvar].values, df[self._yvar].values
        arr_point      = numpy.array([x_arr, y_arr]).T
        z_arr          = df['Efficiency'] 

        grid_z         = spy_grid(arr_point, z_arr, (grid_x, grid_y), method='cubic')
        
        plt.imshow(grid_z.T, extent=(low_x, hig_x, low_y, hig_y), origin='lower')
    #--------------------------
    def _plot_true(self, df):
        arr_xvar         = df[self._xvar].values
        arr_yvar         = df[self._yvar].values
        arr_zvar         = df['Efficiency'].values

        plt.hist2d(x=arr_xvar, y=arr_yvar, weights=arr_zvar, bins=[self._nstep_x, self._nstep_y])
    #--------------------------
    def save_efficiencies(self, version=None):
        if version is None:
            log.error(f'Invalid version: {version}')
            raise
        else:
            self._version = version

        self._initialize()

        jsn_path= f'{self._out_dir}/eff_{self._dset}_{self._trig}.json'

        if os.path.isfile(jsn_path):
            log.warning(f'Output already found, not remaking it: {jsn_path}')
            df = pnd.read_json(jsn_path)
            self._plot_eff(df, interpolate=False)
            self._plot_eff(df, interpolate=True)
            return df

        self._df_dat = self._get_data()

        df_eff = self._get_efficiencies()
        df_eff.to_json(jsn_path, indent=4)

        self._plot_eff(df_eff, interpolate=False)
        self._plot_eff(df_eff, interpolate=True)

        return df_eff
#------------------------------------------------------------
class scale_reader:
    '''
    Class used to read efficiencies under different BDT working points and
    return scale factor
    '''
    #--------------------------
    def __init__(self, wp=None, version=None, dset=None, trig=None):
        '''
        Parameters
        --------------------------
        wp (dict): Stores working point ends for the scan, i.e. {'BDT_cmb' : (0.5, 1.0)} 
        version (str): Version of efficiency file

        dset (str): Dataset, e.g. r1
        trig (str): Trigger, e.g. MTOS
        '''
        self._d_wp        = wp 
        self._dset        = dset
        self._trig        = trig
        self._vers        = version

        self._arr_bdt     = None
        self._arr_eff     = None

        self._cmb_nom     = None
        self._prc_nom     = None

        self._cmb_new     = self._d_wp['BDT_cmb'] 
        self._prc_new     = self._d_wp['BDT_prc'] 
        self._regex       = 'BDT_cmb > ([\.,\d]+) && BDT_prc > ([\.,\d]+)'

        self._initialized = False
    #--------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._read_nominal_wp()
        self._load_data()

        self._initialized = True 
    #--------------------------
    def _read_nominal_wp(self):
        selection = rs.get('bdt', self._trig, q2bin='high', year = 'none')
        mch = re.match(self._regex, selection)
        if not mch:
            log.error(f'Cannot find WP in: {selection}')
            raise

        [cmb_str, prc_str] = mch.groups()

        self._cmb_nom = float(cmb_str)
        self._prc_nom = float(prc_str)

        log.debug(f'Using nominal WP: {self._cmb_nom:.3f}, {self._prc_nom:.3f}')
    #--------------------------
    def _load_data(self):
        file_path = files('extractor_data').joinpath(f'bdt_eff/{self._vers}/eff_{self._dset}_{self._trig}.json')
        if not os.path.isfile(file_path):
            log.error(f'File not found: {file_path}')
            raise

        df     = pnd.read_json(file_path)

        arr_cmb = df.BDT_cmb.values
        arr_prc = df.BDT_prc.values

        self._arr_bdt = numpy.array([arr_cmb, arr_prc]).T
        self._arr_eff = df.Efficiency.values
    #--------------------------
    def _read_eff(self, cmb=None, prc=None):
        [eff] = spy_grid(self._arr_bdt, self._arr_eff, [cmb, prc], method='cubic')

        if math.isnan(eff):
            log.error(f'Found nan efficiency at: {cmb:.3f}, {prc:.3f}')
            raise
        else:
            log.debug(f'{eff:.3f}-> [{cmb:.3f},{prc:.3f}]')

        return eff
    #--------------------------
    def get_scale(self):
        '''
        Returns
        -------------------
        scale (float): Efficiency ratio between WP and nominal.
        '''
        self._initialize()

        eff_nom = self._read_eff(cmb=self._cmb_nom, prc=self._prc_nom)
        eff_new = self._read_eff(cmb=self._cmb_new, prc=self._prc_new)
        scale   = eff_new / eff_nom
        log.debug(f'{scale:.3f}= {eff_new:.3f}/{eff_nom:.3f}')

        return scale
#------------------------------------------------------------




import os
import math
import numpy
import jacobi            as jac
import matplotlib.pyplot as plt

from logzero import logger as log
#----------------------------------------
class combiner:
    '''
    Class used to combine CK values from multiple datasets into
    a combined one
    '''
    #----------------------------------------
    def __init__(self, rjp=None, eff=None, cov=None):
        '''
        rjp (dict): It holds the values of rjpsi e.g. {'2017_TIS' : 1.02...}
        eff (dict): It holds the muon and electron full efficiencies, e.g. {'r1_TOS' : (eff_mu, eff_ee)}
        cov (narray): It's the NxN covariance matrix associated to CK, running over triggers and datasets
        '''

        self._d_rjpsi = rjp
        self._d_eff   = eff
        self._cov     = cov
        self._l_dset  = ['r1', 'r2p1', '2017', '2018']
        self._l_trig  = ['TOS', 'TIS']

        self._out_dir     = None
        self._initialized = False
    #----------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        right_sizes = len(self._d_rjpsi)   == len(self._d_eff) == self._cov.shape[0] == self._cov.shape[1]
        right_keys  = self._d_rjpsi.keys() == self._d_eff.keys()

        if not right_sizes:
            log.error(f'Sizes of inputs is wrong:')
            raise

        if not right_keys:
            log.error(f'Sizes of input dictionaries are different')
            raise

        self._initialized = True
    #----------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot make directory: {value}')
            raise
        self._out_dir = value
    #----------------------------------------
    def _get_ck(self):
        l_ck_tos = []
        l_ck_tis = []
        for dset in self._l_dset:
            for trig in self._l_trig:
                key = f'{dset}_{trig}'
                eff_mm, eff_ee = self._d_eff[key]
                rjpsi          = self._d_rjpsi[key]
                ck             = (eff_ee / eff_mm) / rjpsi
                if trig == 'TOS':
                    l_ck_tos.append(ck)
                else:
                    l_ck_tis.append(ck)

        return numpy.array(l_ck_tos), numpy.array(l_ck_tis)
    #----------------------------------------
    def _get_cov(self, trig=None):
        l_ind = [0, 2, 4, 6] if trig == 'TIS' else [1, 3, 5, 7]

        cov = self._cov
        cov = numpy.delete(cov, l_ind, axis=0)
        cov = numpy.delete(cov, l_ind, axis=1)

        return cov
    #----------------------------------------
    def _plot_avg(self, mu=None, sg=None, arr=None, mat=None, preffix=None):
        if self._out_dir is None:
            return

        l_var=numpy.diag(mat)
        l_err=[math.sqrt(var) for var in l_var]

        plt.errorbar(x=self._l_dset, y=arr, yerr=l_err, label='Measured')
        plt.gca().axhline(y=mu     , linestyle='-', color='r', label='Average')
        plt.gca().axhline(y=mu + sg, linestyle=':', color='r', label=r'$+\sigma$')
        plt.gca().axhline(y=mu - sg, linestyle=':', color='r', label=r'$-\sigma$')
        plt.ylabel(r'$\frac{\varepsilon(ee)}{\varepsilon(\mu\mu)}\cdot\frac{1}{r_{J/\psi}}$')
        plt.tight_layout()
        plt.grid()
        plt.savefig(f'{self._out_dir}/average_{preffix}.png')
        plt.close('all')
    #----------------------------------------
    def _average(self, arr_ck, cov, preffix=None):
        ck_val, ck_var = jac.propagate(lambda x : sum(x)/len(x), arr_ck, cov)
        ck_val, ck_var = float(ck_val), float(ck_var)

        ck_err = math.sqrt(ck_var)
        self._plot_avg(mu=ck_val, sg=ck_err, arr=arr_ck, mat=cov, preffix=preffix)

        return ck_val, ck_var
    #----------------------------------------
    def get_combination(self, add_tis=True):
        '''
        Parameters
        ------------
        add_tis (bool): If true, will return TIS ck and 2x2 covariance, if false, TIS is dropped

        Returns
        ------------
        Tuple with dictionaries with the rjpsi and efficiencies, e.g:

        d_rjpsi, d_eff, cov = cmb.get_cobination()

        the values of the first two dictionaries themselves are not meaningful, but when used
        to calculate ck, they should provide the right combined value
        '''
        self._initialize()

        arr_ck_tos, arr_ck_tis = self._get_ck()
        cov_tos = self._get_cov(trig='TOS') 
        cov_tis = self._get_cov(trig='TIS')

        ck_val_tos, ck_var_tos = self._average(arr_ck_tos, cov_tos, preffix='tos')
        ck_val_tis, ck_var_tis = self._average(arr_ck_tis, cov_tis, preffix='tis')

        eff_val = 1 - 1e-6
        if add_tis:
            d_rjpsi = {'all_TOS' : 1., 'all_TIS' : 1.}
            d_eff   = {'all_TOS' : (eff_val, ck_val_tos), 'all_TIS' : (eff_val, ck_val_tis) }
            cov     = [[ck_var_tos, 0], [0, ck_var_tis]]
        else:
            d_rjpsi = {'all_TOS' : 1.}
            d_eff   = {'all_TOS' : (eff_val, ck_val_tos)}
            cov     = [[ck_var_tos]]

        return d_rjpsi, d_eff, numpy.array(cov)
#----------------------------------------

